#!/usr/bin/env python3
#
# Python script to produce figures for paper "The climate in climate economoics"
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
# Input:  what figures to do [ selecte below via yes / no ]
# Output: the figures, as pngs
#
# To run: exec(open("./Figs4Paper.py").read()) [python3] or %run Figs4Paper.py [ipython]
#
# Dependencies: The script relies on TestDefs.py
#
# Purpose:
# The script is meant to ascertain reproducability of figures used in the paper.
#
# User can choose (yes/no) what pre-defined figures to actually produce.
# Figures are hard wired in this script (Figs4Paper.py) using 'home grown' plotting functions from TestDefs.py
# Tests are hard wired in TestDefs.py
# Test data (input, simulation results, etc) is passed from TestDefs.py to here via a pandas data frame, typically named df2.
#
# To access 'by hand' the simulation data from a particular simulation, returned via a pandas data frame named df2:
#   i=0                         #which simulation to choose? an integer >= 0
#   sisi=df2['simu'][i]         #short hand for the simulation data stored in df2
#   M_AT=sisi.M_of_t[0,:]       #atmospheric mass in the simulation data, in GtC
#   T_AT=sisi.T_of_t[0,:]       #atmospheric temperature change wrt time t=0 in the simulation data, in degree Celsius (or Kelvin, if you prefere)
#   Y_AT=sisi.n_year[:]         #array of years covered by the simulation data
#
#
#
#import libraries we will need
#========================================================================================
#import numpy as np                 #import numpy module, for basic numerics like square roots ( np.sqrt() ) and numpy arrays ( np.arange() )
#import matplotlib.pyplot as plt    #import pyplot from matplotlib module, for basic line plots etc. ( plt.plot(), plt.scatter() )
#import pandas as pd                #import pandas module, for bundeling various data into pandas data frames
#import importlib                   #import importlib module, for re-importing / reloading already imported modules; useful for coding / debugging
import TestDefs as td               #import TestDefs module, 'home grown', defines various tests for Climate Emulator and associated plotting functions
import PatternScaling as ps         #import PatternScaling module, 'home grown', defines methods to scale global mean temperature change to 2D map
#import ClimDICE as cld             #import ClimDICE module, 'home grown', defines class Climate-Carbon-Cycle (methods implementing the CE from DICE)

#this is development software; reload home grown modules using 'importlib' to always have latest edits 'properly loaded'
#importlib.reload(td)    #reload TestDefs.py, to make sure any changes in that file come into effect; useful for coding / debugging
#importlib.reload(ps)    #reload PatternScaling.py, to make sure any changes in that file come into effect; useful for coding / debugging
#importlib.reload(cld)   #reload ClimDICE.py, to make sure any changes in that file come into effect; useful for coding / debugging

import globals                      #specific for IPP, a hack, a 'helper-module' to pass a few variables to IPP specific excercise scripts

#      !!!!!USER DEFINED INPUT STARTS HERE!!!!!
#========================================================================================
#what tests / sets of figures to plot / not plot? ('yes'/'no')


#CMIP5, from carbon emissions to global mean temperature change, RCPs 26 / 45 / 60 / 85
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
F_RCP26E  = 'yes'   #CMIP5, RCP26 from emi.; Figs: temp. / CO2 conc.
F_RCP45E  = 'no'   #CMIP5, RCP45 from emi.; Figs: temp. / CO2 conc.
F_RCP60E  = 'no'   #CMIP5, RCP60 from emi.; Figs: temp. / CO2 conc.
F_RCP85E  = 'no'   #CMIP5, RCP85 from emi.; Figs: temp. / CO2 conc.


#CMIP5, from carbon emissions to 2D temperature change patterns, RCP85
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
PS_RCP85E  = 'yes'; PS_RCP85E_TY=2080   #CMIP5, RCP85 from emi.; temperature pattern scaling for one year


#CMIP5, from CO2 concentrations to global mean temperature change, RCPs 26 / 45 / 60 / 85 (no carbon cycle involved)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
F_RCP26C  = 'no'   #CMIP5, RCP26 from conc.; Figs: temp.
F_RCP45C  = 'no'   #CMIP5, RCP45 from conc.; Figs: temp.
F_RCP60C  = 'no'   #CMIP5, RCP60 from conc.; Figs: temp.
F_RCP85C  = 'no'   #CMIP5, RCP85 from conc.; Figs: temp.


#CMIP5, illustrate emissions, concentrations, forcings
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
F_RCPE    = 'no'  #emissions as input, different hist+RCPs: plot carbon emissions; carbon mass in different reservoirs; forcing
F_RCPC    = 'no'  #CO2 concentrations as input, different hist+RCPs: plot CO2 concentrations; forcing


#Idealized test cases: 4xCO2 (Geoffrey et al. 2013); instantaneous 100 GtC pulse (Joos et al. 2013)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
F_1pctCO2 = 'no'   #CMIP5, 1pctCO2; 1 per cent CO2 increase per year from 1850 onward; Figs: temp.
F_G13     = 'no'   #G13 test; Instantaneous 4xCO2 concentration; Geoffrey et al. 2013; Figs: temp.
F_J13     = 'no'   #J13 test; Instantaneous 100 GtC pulse; Joos et al. 2013; Figs: mass frac. puls / CO2 conc. / temp.


#      !!!!!USER DEFINED INPUT ENDS HERE!!!!!
#========================================================================================




#-----------------------------------------------------------------------------------------
#CMIP5, from carbon emissions to global mean temperature change, RCPs 26 / 45 / 60 / 85
#-----------------------------------------------------------------------------------------
#CMIP5, historical plus RCP26 from carbon emissions
if (F_RCP26E == 'yes'):
    df2=td.BenchMarkTDEmiOrConc(CMIP='RCP26_EMI')
    td.PlotTDTemp(df2,axis_range=[1850.,2100.,-0.5,3.5],avyr=50,fout="CMIPBM_RCP26_EMI_Temp.png", leg='no')
    td.PlotTDMassPPM(df2,axis_range=[1850.,2100.,250.,550.],skip_simu=[1,2,3],cmip_indi='no',pm5pct='yes',fout="CMIPBM_RCP26_EMI_PPM.png", leg='yes')
    #FOR IPP ONLY:
    sisi = df2['simu'][0]
    globals.M_AT = sisi.M_of_t[0,:]
    globals.T_AT = sisi.T_of_t[0,:]
    globals.Y_AT = sisi.n_year[:]

#CMIP5, historical plus RCP45 from carbon emissions
if (F_RCP45E == 'yes'):
    df2=td.BenchMarkTDEmiOrConc(CMIP='RCP45_EMI')
    td.PlotTDTemp(df2,axis_range=[1850.,2100.,-0.5,4.5],avyr=50,fout="CMIPBM_RCP45_EMI_Temp.png", leg='no')
    td.PlotTDMassPPM(df2,axis_range=[1850.,2100.,250.,700.],skip_simu=[1,2,3],cmip_indi='no',pm5pct='yes',fout="CMIPBM_RCP45_EMI_PPM.png", leg='no')

#CMIP5, historical plus RCP60 from carbon emissions
if (F_RCP60E == 'yes'):
    df2=td.BenchMarkTDEmiOrConc(CMIP='RCP60_EMI')
    td.PlotTDTemp(df2,axis_range=[1850.,2100.,-0.5,5.],avyr=50,fout="CMIPBM_RCP60_EMI_Temp.png", leg='no')
    td.PlotTDMassPPM(df2,axis_range=[1850.,2100.,250.,800.],skip_simu=[1,2,3],cmip_indi='no',pm5pct='yes',fout="CMIPBM_RCP60_EMI_PPM.png", leg='no')

#CMIP5, historical plus RCP85 from carbon emissions
if (F_RCP85E == 'yes'):
    df2=td.BenchMarkTDEmiOrConc(CMIP='RCP85_EMI')
    td.PlotTDTemp(df2,axis_range=[1850.,2100.,-0.5,6.],avyr=50,fout="CMIPBM_RCP85_EMI_Temp.png", leg='no')
    td.PlotTDMassPPM(df2,axis_range=[1850.,2100.,250.,1100.],skip_simu=[1,2,3],cmip_indi='no',pm5pct='yes',fout="CMIPBM_RCP85_EMI_PPM.png", leg='no')


    

#-----------------------------------------------------------------------------------------
#CMIP5, from carbon emissions to 2D temperature change patterns, RCP85
#-----------------------------------------------------------------------------------------
#CMIP5, historical plus RCP85 from carbon emissions
if (PS_RCP85E == 'yes'):
    df2=td.BenchMarkTDEmiOrConc(CMIP='RCP85_EMI')   #carry out a CMIP5 RCP simulation with CDICE (simulation 0)
    M4P='MPI-ESM-LR'                                #use CMIP5 model M4P for pattern scaling (global mean temperature change, GMTC, scaled to 2D pattern)
    #M4P='HadGEM2-ES'                                #use CMIP5 model M4P for pattern scaling (global mean temperature change, GMTC, scaled to 2D pattern)
    nys=2080                                        #scale for year nys (determine GMTC between first year of simulation data and year nys)
    lons, lats, T2D_scal, T2D_clim, T2D_patt, GMTC = ps.PatScalTemp(df2, isimu=0, nys=nys, M4P=M4P)
    ps.PlotTempPat(lons, lats, T2D_scal, proj='Mollweide', M4P=M4P, nys=nys) #M4P and nys are used for figure annotation only
    ####ps.PlotTempPat(lons, lats, T2D_scal, proj='InterruptedGoodeHomolosine')
    #FOR IPP ONLY:
    globals.PS_lons     = lons
    globals.PS_lats     = lats
    globals.PS_T2D_scal = T2D_scal
    globals.PS_T2D_clim = T2D_clim
    globals.PS_T2D_patt = T2D_patt
    globals.PS_GMTC     = GMTC



    

#-----------------------------------------------------------------------------------------
#CMIP5, from CO2 concentrations to global mean temperature change, RCPs 26 / 45 / 60 / 85 (no carbon cycle involved)
#-----------------------------------------------------------------------------------------
#CMIP5, historical plus RCP26 from CO2 concentrations
if (F_RCP26C == 'yes'):
    df2=td.BenchMarkTDEmiOrConc(CMIP='RCP26_CONC')
    td.PlotTDTemp(df2,axis_range=[1850.,2100.,-0.5,3.5],avyr=50,fout="CMIPBM_RCP26_CONC_Temp.png", leg='yes')

#CMIP5, historical plus RCP45 from CO2 concentrations
if (F_RCP45C == 'yes'):
    df2=td.BenchMarkTDEmiOrConc(CMIP='RCP45_CONC')
    td.PlotTDTemp(df2,axis_range=[1850.,2100.,-0.5,4.5],avyr=50,fout="CMIPBM_RCP45_CONC_Temp.png", leg='no')

#CMIP5, historical plus RCP60 from CO2 concentrations
if (F_RCP60C == 'yes'):
    df2=td.BenchMarkTDEmiOrConc(CMIP='RCP60_CONC')
    td.PlotTDTemp(df2,axis_range=[1850.,2100.,-0.5,5.],avyr=50,fout="CMIPBM_RCP60_CONC_Temp.png", leg='no')

#CMIP5, historical plus RCP85 from CO2 concentrations
if (F_RCP85C == 'yes'):
    df2=td.BenchMarkTDEmiOrConc(CMIP='RCP85_CONC')
    td.PlotTDTemp(df2,axis_range=[1850.,2100.,-0.5,6.],avyr=50,fout="CMIPBM_RCP85_CONC_Temp.png", leg='no')


    

#-----------------------------------------------------------------------------------------
#CMIP5, illustrate emissions, concentrations, forcings
#-----------------------------------------------------------------------------------------
#CMIP5, historical plus RCP, emissions as input; plot masses in differnt carbon reservoirs, emissions, forcing
if (F_RCPE == 'yes'):
    df_26=td.BenchMarkTDEmiOrConc(CMIP='RCP26_EMI')
    df_45=td.BenchMarkTDEmiOrConc(CMIP='RCP45_EMI')
    df_60=td.BenchMarkTDEmiOrConc(CMIP='RCP60_EMI')
    df_85=td.BenchMarkTDEmiOrConc(CMIP='RCP85_EMI')
    td.PlotResMassGtC(df_26, df_45, df_60, df_85,axis_range=[1850.,2100.,250.,2000.],skip_simu=[0,1,2,4,5,6],fout="CMIP_ResMass_RCP_EMI_GtC_DICE.png", leg='yes')
    td.PlotEmiCMIPGtC(df_26, df_45, df_60, df_85,axis_range=[1850.,2100.,-1.,30.],skip_simu=[0,1,2,4,5,6],fout="CMIP_RCP_EMI_GtC.png", leg='yes')
    td.PlotForcWm2CMIP(df_26, df_45, df_60, df_85,axis_range=[1850.,2100.,0.,8.],skip_simu=[0,1,2,4,5,6],fout="CMIP_RCP_EMI_Forc_Wm2.png", leg='no')

#CMIP5, historical plus RCP, CO2 concentrations as input; plot CO2 concentrations and forcing
if (F_RCPC == 'yes'):
    df_26=td.BenchMarkTDEmiOrConc(CMIP='RCP26_CONC')
    df_45=td.BenchMarkTDEmiOrConc(CMIP='RCP45_CONC')
    df_60=td.BenchMarkTDEmiOrConc(CMIP='RCP60_CONC')
    df_85=td.BenchMarkTDEmiOrConc(CMIP='RCP85_CONC')
    td.PlotConcCMIPppm(df_26, df_45, df_60, df_85,axis_range=[1850.,2100.,250.,950.],skip_simu=[0,1,2,4,5,6],fout="CMIP_RCP_CONC_ppm.png", leg='no')
    td.PlotForcWm2CMIP(df_26, df_45, df_60, df_85,axis_range=[1850.,2100.,0.,8.],skip_simu=[0,1,2,4,5,6],fout="CMIP_RCP_CONC_Forc_Wm2.png", leg='no')



    
#-----------------------------------------------------------------------------------------
#Idealized test case
#-----------------------------------------------------------------------------------------
#CMIP5, 1pct CO2 increase from 1850 onward
if (F_1pctCO2 == 'yes'):
    df2=td.BenchMarkTDEmiOrConc(CMIP='1pctCO2')
    td.PlotTDTemp(df2,axis_range=[0.,70.,-0.2,3.],avyr=10,pad=1.8,fout="CMIPBM_1pctCO2_Temp_70yr.png")
    td.PlotTDTemp(df2,axis_range=[0.,140.,-0.2,6.],avyr=10,pad=1.8,leg='no',fout="CMIPBM_1pctCO2_Temp_140yr.png")
    td.PlotTDMassPPM(df2,axis_range=[0.,140.,250.,1200.],skip_simu=[1,2,3,4,5,6],pm5pct='no',fout="CMIPBM_1pctCO2_PPM.png")

#4xCO2 benchmark from Geoffrey et al. (2013)
if (F_G13 == 'yes'):
    #del df2
    df2 = td.BenchMark4xCO2()
    td.Plot4xCO2Temp(df2,axis_range=[0.,10.,0.,4.0],geoff_yn='yes',geoff_indi='no',leg='yes',fout='TempBM_4xCO2_on_PI_10yr.png')
    td.Plot4xCO2Temp(df2,axis_range=[0.,100.,0.,6.0],geoff_yn='yes',geoff_indi='no',leg='no',fout='TempBM_4xCO2_on_PI_100yr.png')
    td.Plot4xCO2Temp(df2,axis_range=[0.,1000.,0.,9.0],geoff_yn='yes',geoff_indi='no',leg='no',fout='TempBM_4xCO2_on_PI_1000yr.png')

#100GtC puls to PD atmosphere benchmark from  Joos et al. (2013)
if (F_J13 == 'yes'):
    #del df2
    df2 = td.BenchMarkPuls()
    #td.PlotPulsMassPPM(df2, axis_range=[0.,200.,0.,50.], fout='CCBM_100GtC_Puls_PPM.png', joos_yn='yes')
    td.PlotPulsMassFrac(df2, axis_range=[0.,20.,0.2,1.0], fout='CCBM_100GtC_Puls_MassFrac_20yr.png',joos_yn='yes',joos_indi='yes',joos_indi2='no',fsl=12,lsp=0.08,leg='yes')
    td.PlotPulsMassFrac(df2, axis_range=[0.,200.,0.2,1.0], fout='CCBM_100GtC_Puls_MassFrac_200yr.png', joos_yn='yes', joos_indi='yes', joos_indi2='no', leg='no')
    td.PlotPulsMassFrac(df2, axis_range=[0.,1000.,0.2,1.0], fout='CCBM_100GtC_Puls_MassFrac_1000yr.png', joos_yn='yes', joos_indi='yes', joos_indi2='no', leg='no')








    
