#
# To use: import PatternScaling as ps
#
# To run: exec(open("./PatternScaling.py").read())
#
#
# https://www.dkrz.de/up/services/analysis/visualization/sw/pyngl/dkrz-python-example-read-netcdf-file-using-xarray
#
#


import os
import numpy as np
import matplotlib
matplotlib.use("TkAgg")   #needed on my laptop for ipythen to display figures properly
import matplotlib.pyplot as plt
import xarray as xr
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from cartopy.util import add_cyclic_point

#def PatScalTemp(DGMT=2.0, M4P='MPI-ESM-LR', P2P='../pattern_library/')
def PatScalTemp(df2, isimu=0, nys=0, M4P='MPI-ESM-LR', P2P='../pattern_library/', talktalk=1):

        """Pattern Scaling: Global Mean Temperature Change (GMTC) -> 2D Map of Temperature Change (Pattern * GMTC)

        Input, compelling:
        df2      : data frame created by methods in TestDefs.py (BenchMarkPuls, BenchMark4xCO2, BenchMarkTDEmiOrConc)
        isimu    : which simulation (entry) from data frame to use for pattern scaling
        nys      : which year (e.g. 2080) from simulation to use for pattern scaling

        Input, optional:
        M4P      : names of models that provide scaling patterns [list is given below]
        P2P      : relative path to files containing scaling patterns (default: ../pattern_library)
        talktalk : talk a bit more upon execution if set to 1

        List of models for pattern scaling (all based on RCP85; Lynch et al., Earth System Science Data, Vol. 9, Number 1, 5 2017):
        ACCESS1-0, ACCESS1-3, BNU-ESM, CanESM2, CCSM4, CESM1-BGC, CESM1-CAM5, CESM1-WACCM, CMCC-CESM, CMCC-CM, CMCC-CMS, 
        CNRM-CM5, CSIRO-Mk3-6-0, EC-EARTH, FGOALS-g2, FIO-ESM, GFDL-CM3, GFDL-ESM2G, GFDL-ESM2M, GISS-E2-H-CC, GISS-E2-H, 
        GISS-E2-R-CC, GISS-E2-R, HadGEM2-AO, HadGEM2-CC, HadGEM2-ES, IPSL-CM5A-LR, IPSL-CM5A-MR, IPSL-CM5B-LR, MIROC5, 
        MIROC-ESM-CHEM, MIROC-ESM, MPI-ESM-LR, MPI-ESM-MR, MRI-CGCM3, MRI-ESM1, NorESM1-ME, NorESM1-M, 
        """

        #desired input file for pattern scaling; existing? yes -> load data; no -> tell user and return
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
        #fname   = '../pattern_library/PATTERN_tas_ANN_MPI-ESM-LR_rcp85.nc'
        fname    = P2P+'PATTERN_tas_ANN_'+M4P+'_rcp85.nc'
        lons     = -1.   #default value, in case no pattern file is found
        lats     = -1.   #default value, in case no pattern file is found
        T2D_scal = -1.   #default value, in case no pattern file is found

        #pattern scaling files does not exist: tell user and return
        if not os.path.isfile(fname):
            print("ERROR! File for pattern scaling not found! File is "+fname)
            return lons, lats, T2D_scal

        #pattern scaling file exists: open file (using xarray) and read what is need
        ds       = xr.open_dataset(fname)
        T2D_patt = ds.pattern
        T2D_clim = ds.climatology
        lons     = ds.lon
        lats     = ds.lat
        #pattern data needs adding one set of longitudes, to 'close' the map at longitud = 0 degrees
        T2D_patt, lons = add_cyclic_point(T2D_patt, axis=1, coord=lons)  #tell python / cartopy to add a cylclic entry to longitudes, to get rid of white line at lon=0
        T2D_clim       = add_cyclic_point(T2D_clim, axis=1)              #tell python / cartopy to add a cylclic entry to longitudes, to get rid of white line at lon=0


        #desired simulation and target year: existing?
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        nsimu = len(df2['simu'])
        #desired simulation does not exist: tell user and return
        if (isimu > nsimu-1):
            print("ERROR! Desired simulation isimu = "+ str(isimu) +" exceeds maximum index "+ str(nsimu-1) +" of available simulations")
            return lons, lats, T2D_scal

        #desired simulation does exist: introduce short hand (sisi = ...) to it
        sisi = df2['simu'][isimu]
        
        #get index of target year nys from array of available simulation years (default: last year available)
        idx  = np.where(sisi.n_year == nys)  #for np.where(), see e.g. https://thispointer.com/find-the-index-of-a-value-in-numpy-array/
        #desired year not found: use last available year as target year
        if not all(idx):
            ityr = len(sisi.n_year)-1
            print("WARNING! Desired year not found, using last simulation year available, i.e., year "+ str(sisi.n_year[ityr]))
        #desired target year found
        else:
            ityr = idx[0][0]


        #get global mean temperature change DGTM (desired year minus first year) and use it for scaling, yielding T2D_scal
        GMTC     = sisi.T_of_t[0,ityr] - sisi.T_of_t[0,0] 
        #T2D_scal = T2D_clim + GMTC*T2D_patt   #annual mean temperature in degree Celsius
        T2D_scal = GMTC*T2D_patt               #change in annual mean temperature in degree Celsius

        if (talktalk==1):
            print("pattern scaling:")
            print("base and target year are %4i and %4i" % (sisi.n_year[0], sisi.n_year[ityr]))   #formatted output of integers
            print("global mean temperature change is %5.2f degrees" % (GMTC))                     #formatted output of floats

        return lons, lats, T2D_scal, T2D_clim, T2D_patt, GMTC


    
def PlotTempPat(lons, lats, T2D_scal, levels=60, proj='PlateCarree', addcont='no', addcb='yes', cmap='Reds', figx=10, figy=6, \
                fout='TemperatureChangeMap.png', M4P='MPI-ESM-LR', nys=2080):


        """Pattern Scaling: Plot Map of (Scaled) 2D Temperature Field

        Data is assumed to come as Plate Carree (or Equirectangular form; equidistant in longitude and latidue).

        Input, compelling:
        lons, lats, T2D_scal: longitudes (1D array), latitudes (1D array), data to plot (2D array of lon/lat)

        Input, optional:
        proj    : projections for plotting [PlateCarree, Mercator, Miller, Mollweide, Robinson, InterruptedGoodeHomolosine, RotatedPole]
        levels  : levels of filled (color) contours (number of levels if only one number is passed)
        addcont : add contours to plot 
        addcb   : add a color bar
        cmap    : colormap to use [default 'Reds'; alternatives e.g. 'jet']
        figx    : figure size in inches, horizontal
        figy    : figure size in inches, vertical
        fout    : name of output file (png)

        Output: 
        geographical map (2D plot)
        """
        #to see the above doc on a python3 command prompt: import PatternScaling as ps; print(ps.PlotTempPat.__doc__)
        #for colormaps: https://matplotlib.org/stable/tutorials/colors/colormaps.html

        #what projection?
        data_crs=ccrs.PlateCarree()  #default projection
        proj_crs=ccrs.PlateCarree()
        if (proj=='PlateCarree'):                  proj_crs=ccrs.PlateCarree()
        elif (proj=='Mercator'):                   proj_crs=ccrs.Mercator()
        elif (proj=='Miller'):                     proj_crs=ccrs.Miller()
        elif (proj=='Mollweide'):                  proj_crs=ccrs.Mollweide()
        elif (proj=='Robinson'):                   proj_crs=ccrs.Robinson()
        elif (proj=='InterruptedGoodeHomolosine'): proj_crs=ccrs.InterruptedGoodeHomolosine()
        elif (proj=='RotatedPole'):                proj_crs=ccrs.RotatedPole()
        else: print("WARNING! Desired projection "+ proj +" not found, using PlateCarree instead")
        print("projection is "+ proj)
        
        fig = plt.figure(figsize=(figx, figy))
        ax = fig.add_subplot(1, 1, 1, projection=proj_crs)
        ax.set_global()
        ax.coastlines()
        cf = ax.contourf(lons, lats, T2D_scal, levels=levels, cmap=cmap, transform=data_crs) #plot filled contours, 60 levels
        if (addcont=='yes'): cl = ax.contour(lons, lats, T2D_scal, levels=cf.levels, colors=['black'], transform=data_crs) #add black contour lines
        density=4
        ax.contourf(lons, lats, T2D_scal, transform=data_crs, colors='none', levels=[4.0, 20.0], hatches=[density*'/',density*'/'],)
        #ZHlon=8.54; ZHlat=47.38; ax.scatter(ZHlon, ZHlat, transform=ccrs.PlateCarree())
        #ax.gridlines()
        #ax.add_feature(cfeature.OCEAN)
        #ax.add_feature(cfeature.LAND)
        #ax.add_feature(cfeature.LAKES)
        #ax.add_feature(cfeature.BORDERS)
        #ax.add_feature(cfeature.COASTLINE)
        #ax.add_feature(cfeature.RIVERS)
        if (addcb=='yes'): cb = plt.colorbar(cf, shrink=0.5)
        cb.ax.set_title('$^{o}$Celsius')
        ax.set_title('Temperature Change [model / year : '+M4P+' / '+str(nys)+']')
        plt.tight_layout(); 
        plt.show(block=False)
        plt.savefig(fout,format='png')
